"""
Copyright 2015 Ingo Fruend (github@ingofruend.net)
"""
from __future__ import (division, absolute_import, print_function,
                        unicode_literals)

from optparse import OptionParser

import beets
from beets.library import LibModel, Item, Album, Library

from beets.plugins import BeetsPlugin
from beets.ui import Subcommand
from beets.dbcore import types

_DELIMITER = '|'

class UserTagsPlugin(BeetsPlugin):
    """UserTags plugin to support user defined tags"""
    FIELD = 'usertags'
    item_types = {'usertags': types.STRING}
    album_types = {'usertags': types.STRING}

    def __init__(self):
        super(UserTagsPlugin, self).__init__()
        self._addtag_cmd = self._create_add_command()
        self._rmtag_cmd = self._create_remove_command()
        self._cleartags_cmd = self._create_clear_command()
        self._listtags_cmd = self._create_list_command()
        self.config.add({'auto': False, 'album_tags': None, 'item_tags': None})
        self.register_listener('album_imported', self._on_album_imported)
        self.register_listener('item_imported', self._on_item_imported)

    def commands(self):
        return [self._addtag_cmd,
                self._rmtag_cmd,
                self._cleartags_cmd,
                self._listtags_cmd]

    @staticmethod
    def get_tags(model: LibModel) -> list[str]:
        if isinstance(model, Item):
            tags = model.get(UserTagsPlugin.FIELD, default=None, with_album=False)
        elif isinstance(model, Album):
            tags = model.get(UserTagsPlugin.FIELD, None)
        else:
            tags = None
        return tags.split(_DELIMITER) if tags else []

    def add_tags(self, lib, opts, args):
        new_tags = self._sanitize_tags(opts.tags or [])
        if not new_tags:
            self._log.error("Please specify at least one valid tag to add!\n")
            self._addtag_cmd.print_help()
            return

        models = self._get_models(lib, opts.album, args)
        if not self._check_models(models, opts.album): return

        if not self._prompt_if_required(
                opts.prompt, models,
                prompt_text="This will add the tag(s) {} to the following {}:"
                        .format(', '.join(new_tags), "album(s)" if opts.album else "track(s)"),
                default_text="Adding tag(s) {} to:".format(', '.join(new_tags))):
            return

        for model in models:
            self._add_tags(model, new_tags)
            if not opts.prompt: self._log.info("  {}".format(model))
            if opts.inherit and isinstance(model, Album):
                self._log.info(f"Adding tags {new_tags} to items of {model}")
                for item in model.items():
                    self._add_tags(item, new_tags)

    def remove_tags(self, lib, opts, args):
        remove_tags = self._sanitize_tags(opts.tags or [])
        if not remove_tags:
            self._log.warn("Please specify at least one valid tag to remove!\n")
            self._rmtag_cmd.print_help()
            return

        models = self._get_models(lib, opts.album, args)
        if not self._check_models(models, opts.album): return

        if not self._prompt_if_required(
                opts.prompt, models,
                prompt_text="This will remove the tag(s) {} from the following {}:"
                        .format(', '.join(remove_tags), "album(s)" if opts.album else "track(s)"),
                default_text="Removing tag(s) {} from:".format(', '.join(remove_tags))):
            return

        for model in models:
            self._remove_tags(model, remove_tags)
            if not opts.prompt: self._log.info('  {}'.format(model))
            if opts.inherit and isinstance(model, Album):
                self._log.info(f"Removing tags {remove_tags} from items of {model}")
                for item in model.items():
                    self._remove_tags(item, remove_tags)

    def clear_tags(self, lib, opts, args):
        models = self._get_models(lib, opts.album, args)
        if not self._check_models(models, opts.album): return

        if not self._prompt_if_required(
                opts.prompt, models,
                prompt_text="This will remove ALL tags from the following {}:"
                        .format("album(s)" if opts.album else "track(s)"),
                default_text="Removing ALL tags from:"):
            return

        for model in models:
            model.update({UserTagsPlugin.FIELD: None})
            self._update_model(model)
            if not opts.prompt: self._log.info("  {}".format(model))

    def list_tags(self, lib, opts, args):
        models = self._get_models(lib, opts.album, args)
        if not self._check_models(models, opts.album): return

        tags = []
        for model in models:
            tags += self.get_tags(model)
        for tag in sorted(set(tags)):
            print(tag, len([True for t in tags if t == tag]))

    def _create_add_command(self):
        cmd = Subcommand(
            'addtag',
            help='add user-defined tags',
            aliases=('adt',))
        cmd.func = self.add_tags
        self._add_tag_option(cmd.parser)
        self._add_prompt_option(cmd.parser)
        self._add_inherit_option(cmd.parser)
        cmd.parser.add_album_option()
        return cmd

    def _create_remove_command(self):
        cmd = Subcommand(
            'rmtag',
            help='remove user-defined tags',
            aliases=('rmt',))
        cmd.func = self.remove_tags
        self._add_tag_option(cmd.parser)
        self._add_prompt_option(cmd.parser)
        self._add_inherit_option(cmd.parser)
        cmd.parser.add_album_option()
        return cmd

    def _create_clear_command(self):
        cmd = Subcommand(
            'cleartags',
            help='remove ALL user-defined tags from tracks')
        cmd.func = self.clear_tags
        self._add_prompt_option(cmd.parser)
        cmd.parser.add_album_option()
        return cmd

    def _create_list_command(self):
        cmd = Subcommand(
            'listtags',
            help='list all user-defined tags on tracks',
            aliases=('lst',))
        cmd.func = self.list_tags
        cmd.parser.add_album_option()
        return cmd

    def _on_album_imported(self, lib: Library, album: Album):
        if not self.config['auto']:
            return
        album_tags = self._sanitize_tags(self.config['album_tags'].as_str_seq())
        if album_tags:
            self._add_tags(album, album_tags)
            self._log.debug("Added tag(s) {} to album on import: {}".format(album_tags, album))
        item_tags = self._sanitize_tags(self.config['item_tags'].as_str_seq())
        if item_tags:
            for item in album.items():
                self._add_tags(item, item_tags)
                self._log.debug("Added tag(s) {} to item on import: {}".format(item_tags, item))

    def _on_item_imported(self, lib: Library, item: Item):
        if not self.config['auto']:
            return
        item_tags = self._sanitize_tags(self.config['item_tags'].as_str_seq())
        if item_tags:
            self._add_tags(item, item_tags)
            self._log.debug("Added tag(s) {} to item on import: {}".format(item_tags, item))

    def _add_tags(self, model: LibModel, new_tags: list[str]):
        tags = self.get_tags(model)
        tags.extend(new_tags)
        tags = sorted(list(set(tags)))
        model.update({UserTagsPlugin.FIELD: _DELIMITER.join(tags)})
        self._update_model(model)

    def _remove_tags(self, model, remove_tags):
        tags = self.get_tags(model)
        tags = [tag for tag in tags if tag not in remove_tags]
        tags_field = _DELIMITER.join(tags) if tags else None
        model.update({UserTagsPlugin.FIELD: tags_field})
        self._update_model(model)

    @staticmethod
    def _add_tag_option(parser: OptionParser):
        parser.add_option(
            '--tag', '-t',
            action='append', dest='tags',
            help='tag to add/remove; one tag per flag')

    @staticmethod
    def _add_prompt_option(parser: OptionParser):
        parser.add_option(
            '--prompt', '-p',
            action='store_true', default=False,
            dest='prompt', help='prompt user for confirmation before making changes'
        )

    @staticmethod
    def _add_inherit_option(parser: OptionParser):
        parser.add_option(
            '--inherit', '-i',
            action='store_true', default=False,
            dest='inherit', help='when changing album tags, make the same changes for its items'
        )

    @staticmethod
    def _get_models(lib: Library, album: bool, args: list[str]) -> list[LibModel]:
        if album:
            return list(lib.albums(args))
        else:
            return list(lib.items(args))

    def _check_models(self, models: list[LibModel], album: bool) -> bool:
        if not models:
            self._log.info("Query returned no {}".format("albums" if album else "tracks"))
            return False
        else:
            return True

    @staticmethod
    def _update_model(model: LibModel) -> None:
        if isinstance(model, Item):
            model.store()
        elif isinstance(model, Album):
            model.store(inherit=False)

    @staticmethod
    def _sanitize_tags(tags: list[str]) -> list[str]:
        return [tag for tag in tags if UserTagsPlugin._is_tag_valid(tag)]

    @staticmethod
    def _is_tag_valid(tag: str) -> bool:
        return bool(tag.strip())

    def _prompt_if_required(self, prompt: bool, models: list[LibModel], prompt_text: str, default_text: str) -> bool:
        if prompt:
            self._log.info(prompt_text)
            for model in models:
                self._log.info("  {}".format(model))
            if not beets.ui.input_yn("Continue? (Y/n)"):
                self._log.info("No changes made.")
                return False
        else:
            self._log.info(default_text)
        return True


import unittest

import beets

from beetsplug.usertag import UserTagsPlugin
from beets.library import LibModel, Library
from beets.test.helper import TestHelper
from optparse import Values
from unittest.mock import patch

_ITEM_TAG = 'item_tag'
_ALBUM_TAG = 'album_tag'

def _create_opts(album: bool, tags: list[str], prompt: bool=False, inherit: bool=False) -> Values:
    return Values({'album': album, 'tags': tags, 'prompt': prompt, 'inherit': inherit})

_ITEM_OPTS = _create_opts(album=False, tags=[_ITEM_TAG])
_ITEM_OPTS_PROMPT = _create_opts(album=False, tags=[_ITEM_TAG], prompt=True)
_ALBUM_OPTS = _create_opts(album=True, tags=[_ALBUM_TAG])

def with_item_tags(*tags: str):
    def decorator(func):
        func._item_tags = list(tags)
        return func
    return decorator

def with_album_tags(*tags: str):
    def decorator(func):
        func._album_tags = list(tags)
        return func
    return decorator

class UserTagsTest(TestHelper, unittest.TestCase):

    def setUp(self):
        super().setup_beets()
        self.subject = UserTagsPlugin()
        self._create_items()
        test_method = getattr(self, self._testMethodName)
        item_tags = getattr(test_method, "_item_tags", None)
        if item_tags:
            self.subject._add_tags(self.item, item_tags)
        album_tags = getattr(test_method, "_album_tags", None)
        if album_tags:
            self.subject._add_tags(self.album, album_tags)

    # region Add

    def test_adding_tag_item(self):
        self.subject.add_tags(self.lib, _ITEM_OPTS, self.item.title)
        self._assert_item_tags(expected=[_ITEM_TAG])

    def test_adding_tag_album(self):
        self.subject.add_tags(self.lib, _ALBUM_OPTS, self.album.album)
        self._assert_album_tags(expected=[_ALBUM_TAG])

    def test_adding_tag_to_item_does_not_change_album(self):
        self.subject.add_tags(self.lib, _ITEM_OPTS, self.item.title)
        self._assert_item_tags(expected=[_ITEM_TAG])
        self._assert_album_tags(expected=[])

    def test_adding_tag_to_album_does_not_change_item(self):
        self.subject.add_tags(self.lib, _ALBUM_OPTS, self.album.album)
        self._assert_item_tags(expected=[])
        self._assert_album_tags(expected=[_ALBUM_TAG])

    def test_adding_tag_to_album_updates_items_when_inherit(self):
        opts = _create_opts(album=True, tags=[_ALBUM_TAG], inherit=True)
        self.subject.add_tags(self.lib, opts, self.album.album)
        self._assert_item_tags(expected=[_ALBUM_TAG])
        self._assert_album_tags(expected=[_ALBUM_TAG])

    @with_item_tags('baa', 'bab')
    def test_adding_tag_to_album_with_inherit_does_not_change_other_tags_on_item(self):
        opts = _create_opts(album=True, tags=[_ALBUM_TAG], inherit=True)
        self.subject.add_tags(self.lib, opts, self.album.album)
        self._assert_item_tags(expected=[_ALBUM_TAG, 'baa', 'bab'])
        self._assert_album_tags(expected=[_ALBUM_TAG])

    def test_adding_tag_item_multiple_times(self):
        self.subject.add_tags(self.lib, _ITEM_OPTS, self.item.title)
        self._assert_item_tags(expected=[_ITEM_TAG])

        self.subject.remove_tags(self.lib, _ITEM_OPTS, self.item.title)

        self.subject.add_tags(self.lib, _ITEM_OPTS, self.item.title)
        self._assert_item_tags(expected=[_ITEM_TAG])

    def test_invalid_tags_are_stripped_when_adding(self):
        item_opts = _create_opts(
            album=False, tags=['baa', '', 'bab', ' ', 'bac', '   ', 'bad', '\t', 'bae', '	', 'baf'])
        self.subject.add_tags(self.lib, item_opts, self.item.title)
        self._assert_item_tags(expected=['baa', 'bab', 'bac', 'bad', 'bae', 'baf'])

    def test_repeated_tags_are_ignored(self):
        item_opts = _create_opts(album=False, tags=['baa', 'bab', 'baa', 'bac', 'baa'])
        self.subject.add_tags(self.lib, item_opts, self.item.title)
        self._assert_item_tags(expected=['baa', 'bab', 'bac'])

    @with_item_tags('baa', 'bab', 'bac')
    def test_adding_existing_tags(self):
        item_opts = _create_opts(album=False, tags=['baa', 'bad', 'bae', 'bab'])
        self.subject.add_tags(self.lib, item_opts, self.item.title)
        self._assert_item_tags(expected=['baa', 'bab', 'bac', 'bad', 'bae'])

    @patch('beets.ui.input_yn', return_value=True)
    def test_adding_prompt_yes(self, mock):
        self.subject.add_tags(self.lib, _ITEM_OPTS_PROMPT, self.item.title)
        self._assert_item_tags(expected=[_ITEM_TAG])

    @patch('beets.ui.input_yn', return_value=False)
    def test_adding_prompt_no(self, mock):
        self.subject.add_tags(self.lib, _ITEM_OPTS_PROMPT, self.item.title)
        self._assert_item_tags(expected=[])

    @patch.object(Library, 'items')
    def test_check_valid_tags_when_adding(self, mock_items_query):
        item_opts = _create_opts(album=False, tags=['', ' ', '   ', '\t', '	'])
        self.subject.add_tags(self.lib, item_opts, self.item.title)
        mock_items_query.assert_not_called()

    def test_adding_tag_twice(self):
        self.subject.add_tags(self.lib, _ITEM_OPTS, self.item.title)
        self._assert_item_tags(expected=[_ITEM_TAG])

        self.subject.add_tags(self.lib, _ITEM_OPTS, self.item.title)
        self._assert_item_tags(expected=[_ITEM_TAG])

    # endregion

    # region Remove

    @with_item_tags(_ITEM_TAG)
    def test_removing_tag_item(self):
        self.subject.remove_tags(self.lib, _ITEM_OPTS, self.item.title)
        self._assert_item_tags(expected=[])

    @with_album_tags(_ALBUM_TAG)
    def test_removing_tag_album(self):
        self.subject.remove_tags(self.lib, _ALBUM_OPTS, self.album.album)
        self._assert_album_tags(expected=[])

    @with_item_tags('foo')
    @with_album_tags('foo')
    def test_removing_item_tag_does_not_change_album(self):
        item_opts = _create_opts(album=False, tags=['foo'])
        self.subject.remove_tags(self.lib, item_opts, self.item.title)

        self._assert_item_tags(expected=[])
        self._assert_album_tags(expected=['foo'])

    @with_item_tags('foo')
    @with_album_tags('foo')
    def test_removing_album_tag_does_not_change_item(self):
        album_opts = _create_opts(album=True, tags=['foo'])
        self.subject.remove_tags(self.lib, album_opts, self.album.album)

        self._assert_item_tags(expected=['foo'])
        self._assert_album_tags(expected=[])

    @with_item_tags('baa', 'bab')
    @with_album_tags('baa', 'bab')
    def test_removing_tag_from_album_updates_items_when_inherit(self):
        opts = _create_opts(album=True, tags=['baa', 'bab'], inherit=True)
        self.subject.remove_tags(self.lib, opts, self.album.album)
        self._assert_item_tags(expected=[])
        self._assert_album_tags(expected=[])

    @with_item_tags('baa', 'bab', 'bac')
    @with_album_tags('baa', 'bab')
    def test_removing_tag_from_album_with_inherit_does_not_change_other_tags_on_item(self):
        opts = _create_opts(album=True, tags=['bab'], inherit=True)
        self.subject.remove_tags(self.lib, opts, self.album.album)
        self._assert_item_tags(expected=['baa', 'bac'])
        self._assert_album_tags(expected=['baa'])

    @with_item_tags('baa', 'bab', 'bac', 'bad')
    def test_removing_subset(self):
        item_opts = _create_opts(album=False, tags=['baa', 'bac'])
        self.subject.remove_tags(self.lib, item_opts, self.item.title)

        self._assert_item_tags(expected=['bab', 'bad'])

    @with_item_tags('baa', 'bab', 'bac', 'bad', 'bae', 'baf')
    def test_invalid_tags_are_stripped_when_removing(self):
        item_opts = _create_opts(
            album=False,
            tags=['baa', '', 'bab', ' ', 'bac', '   ', 'bad', '\t', 'bae', '	', 'baf'])
        self.subject.remove_tags(self.lib, item_opts, self.item.title)

        self._assert_item_tags(expected=[])

    @with_item_tags(_ITEM_TAG)
    @patch('beets.ui.input_yn', return_value=True)
    def test_removing_prompt_yes(self, mock):
        self.subject.remove_tags(self.lib, _ITEM_OPTS_PROMPT, self.item.title)
        self._assert_item_tags(expected=[])

    @with_item_tags(_ITEM_TAG)
    @patch('beets.ui.input_yn', return_value=False)
    def test_removing_prompt_no(self, mock):
        self.subject.remove_tags(self.lib, _ITEM_OPTS_PROMPT, self.item.title)
        self._assert_item_tags(expected=[_ITEM_TAG])

    @patch.object(Library, 'items')
    def test_check_valid_tags_when_removing(self, mock_items_query):
        item_opts = _create_opts(album=False, tags=['', ' ', '   ', '\t', '	'])
        self.subject.remove_tags(self.lib, item_opts, self.item.title)
        mock_items_query.assert_not_called()

    # endregion

    # region Clear

    @with_item_tags('foo', 'bar')
    def test_clearing_tags_item(self):
        clear_opts = _create_opts(album=False, tags=[])
        self.subject.clear_tags(self.lib, clear_opts, self.item.title)
        self._assert_item_tags(expected=[])

    @with_album_tags('foo', 'bar')
    def test_clearing_tags_album(self):
        clear_opts = _create_opts(album=True, tags=[])
        self.subject.clear_tags(self.lib, clear_opts, self.album.album)
        self._assert_album_tags(expected=[])

    @with_item_tags('foo', 'bar')
    @with_album_tags('foo', 'bar')
    def test_clearing_item_tags_does_not_change_album(self):
        clear_opts = _create_opts(album=False, tags=[])
        self.subject.clear_tags(self.lib, clear_opts, self.item.title)
        self._assert_item_tags(expected=[])
        self._assert_album_tags(expected=['bar', 'foo'])

    @with_item_tags('foo', 'bar')
    @with_album_tags('foo', 'bar')
    def test_clearing_album_tags_does_not_change_item(self):
        clear_opts = _create_opts(album=True, tags=[])
        self.subject.clear_tags(self.lib, clear_opts, self.album.album)
        self._assert_item_tags(expected=['bar', 'foo'])
        self._assert_album_tags(expected=[])

    @with_item_tags('foo', 'bar')
    @patch('beets.ui.input_yn', return_value=True)
    def test_clearing_prompt_yes(self, mock):
        clear_opts = _create_opts(album=False, tags=[], prompt=True)
        self.subject.clear_tags(self.lib, clear_opts, self.item.title)
        self._assert_item_tags(expected=[])

    @with_item_tags('foo', 'bar')
    @patch('beets.ui.input_yn', return_value=False)
    def test_clearing_prompt_no(self, mock):
        clear_opts = _create_opts(album=False, tags=[], prompt=True)
        self.subject.clear_tags(self.lib, clear_opts, self.item.title)
        self._assert_item_tags(expected=['bar', 'foo'])

    # endregion

    # region Import

    def test_album_tags_added_on_import(self):
        self._init_config(auto = True, album_tags = ['baa', 'bab'], item_tags = ['bac', 'bad'])
        self.subject._on_album_imported(self.lib, self.album)

        self._assert_album_tags(expected=['baa', 'bab'])

    def test_album_tags_not_added_on_import(self):
        self._init_config(auto = False, album_tags = ['foo'])
        self.subject._on_album_imported(self.lib, self.album)

        self._assert_album_tags(expected=[])

    def test_item_tags_added_on_import(self):
        self._init_config(auto = True, album_tags = ['baa', 'bab'], item_tags = ['bac', 'bad'])
        self.subject._on_item_imported(self.lib, self.item)

        self._assert_item_tags(expected=['bac', 'bad'])

    def test_item_tags_not_added_on_import(self):
        self._init_config(auto=False, item_tags=['foo'])
        self.subject._on_item_imported(self.lib, self.item)

        self._assert_item_tags(expected=[])

    def test_item_tags_added_when_album_imported(self):
        self._init_config(auto = True, album_tags = ['baa', 'bab'], item_tags = ['bac', 'bad'])
        self.subject._on_album_imported(self.lib, self.album)

        self._assert_item_tags(expected=['bac', 'bad'])

    # endregion

    @staticmethod
    def _init_config(auto: bool, album_tags: list[str] = None, item_tags: list[str] = None):
        beets.config['usertag'].set({'auto': auto, 'album_tags': album_tags or [], 'item_tags': item_tags or []})

    def _create_items(self):
        self.item = self.add_item()
        self.album = self.lib.add_album([self.item])

    def _assert_item_tags(self, expected: list[str]):
        item = self.lib.get_item(self.item.id)
        self._assert_user_tags(item, expected)

    def _assert_album_tags(self, expected: list[str]):
        album = self.lib.get_album(self.album.id)
        self._assert_user_tags(album, expected)

    def _assert_user_tags(self, model: LibModel, expected: list[str]):
        self.assertEqual(expected, UserTagsPlugin.get_tags(model))